# Configuration parser for Beam                -*- perl -*-
# Copyright (C) 2016, 2017 Sergey Poznyakoff <gray@gnu.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

package App::Beam::Config;

use strict;
use Carp;
use File::stat;
use Storable qw(retrieve store);
use App::Beam::Config::Locus;
use Data::Dumper;

require Exporter;
our @ISA = qw(Exporter);

our $VERSION = "1.00";

=head1 NAME

App::Beam::Config - generalized configuration file parser

=head1 SYNOPSIS

    my $cfg = new App::Beam::Config($filename, %opts);
    $cfg->parse() or die;

    if ($cfg->isset('core', 'variable')) {
       ...
    }

    my $x = $cfg->get('file', 'locking');

    $cfg->set('file', 'locking', 'true');

    $cfg->unset('file', 'locking');

=head1 DESCRIPTION

=cut

=head2 $cfg = new App::Beam::Config($filename, %opts);

Creates new configuration object for file B<$filename>.  Valid
options are:

=over 4

=item B<debug> => I<NUM>

Sets debug verbosity level.    

=item B<ci> => B<0> | B<1>

If B<1>, enables case-insensitive keyword matching.  Default is B<0>,
i.e. the keywords are case-sensitive.    

=item B<locations> => B<0> | B<1>

Controls the I<locations> mode.  When locations mode is enabled. the
resulting object will store, along with each configuration setting, the
location in the source file where it was set.

=item B<parameters> => \%hash

Defines the syntax table.  See below for a description of B<%hash>.

=item B<cachefile> => I<FILENAME>

Sets the location of the cache file.  If passed, the parsed configuration
will be stored in binary form in the I<FILENAME>.  Before parsing the
configuration file, the constructor will chech if the cache file exists and
has the same timestamp as the configuration file.  If so, the configuration
will be loaded from the cache (using B<Storable>(3)), avoiding parsing
overhead.  Otherwise, the cached data will be discarded, and the source file
will be parsed as usual.

The destructor will first check if the configuration was updated, and if
so will recreate the cache file prior to destructing the object instance.    

=item B<rw> => B<0> | B<1>

Whether or not the configuration is read-write.  This setting is in effect
only if B<cachefile> is also set.
    
If set to B<0> (the default) any local changes to the configuration (using
B<set> and B<unset> methods), will not be saved to the cache file upon
exiting.  Otherwise, the eventual modifications will be stored in the cache.    
    
=back    

=head3 Syntax hash

The hash passed via the B<parameters> keyword defines the keywords and
sections allowed within a configuration file.  In a simplest case, a
keyword is described as

    name => 1

This means that B<name> is a valid keyword, but does not imply anything
more about it or its value.  A most complex declaration is possible, in
which the value is a hash reference, containing on or more of the following
keywords:

=over 4

=item mandatory => 0 | 1

Whether or not this setting is mandatory.

=item array => 0 | 1

If B<1>, the value of the setting is an array.  Each subsequent occurrence
of the statement appends its value to the end of the array.

=item re => I<regexp>

Defines a regular expression to which must be matched by the value of the
setting, otherwise a syntax error will be reported.

=item check => I<coderef>

Defines a code which will be called after parsing the statement in order to
verify its value.  The I<coderef> is called as

    $err = &{$coderef}($valref, $prev_value)

where B<$valref> is a reference to its value, and B<$prev_value> is the value
of the previous instance of this setting.  The function must return B<undef>
if the value is OK for that setting.  In that case, it is allowed to modify
the value, referenced by B<$varlref>.  If the value is erroneous, the function
must return a textual error message, which will be printed using B<$cfg->error>.
    
=back    

To define a section, use the B<section> keyword, e.g.:

    core => {
        section => {
            pidfile => {
               mandatory => 1
            },
            verbose => {
               re => qr/^(?:on|off)/i
            }
        }
    }

This says that a section B<[core]> can have two variables: B<pidfile>, which
is mandatory, and B<verbose>, whose value must be B<on>, or B<off> (case-    
insensitive).

To allow for arbitrary keywords, use B<*>.  For example, the following
declares the B<[code]> section, which must have the B<pidfile> setting
and is allowed to have any other settings as well.    
 
    code => {
       section => {
           pidfile => { mandatory => 1 },
           '*' => 1
       }
    }
    
=cut

sub new {
    my $class = shift;
    my $filename = shift;
    local %_ = @_;
    my $self = bless { filename => $filename }, $class;
    my $v;
    my $err;
    
    if (defined($v = delete $_{debug})) {
	$self->{debug} = $v;
    }

    if (defined($v = delete $_{ci})) {
	$self->{ci} = $v;
    }

    if (defined($v = delete $_{locations})) {
	$self->{locations} = $v;
    }
    
    if (defined($v = delete $_{parameters})) {
	if (ref($v) eq 'HASH') {
	    $self->{parameters} = $v;
	} else {
	    carp "parameters must refer to a HASH";
	    ++$err;
	}
    }

    if (defined($v = delete $_{cachefile})) {
	$self->{cachefile} = $v;
    }

    if (defined($v = delete $_{cache})) {
	unless (exists($self->{cachefile})) {
	    $v = $self->{filename};
	    $v =~ s/\.(conf|cnf|cfg)$//;
	    unless ($v =~ s#(.+/)?(.+)#$1.$2#) {
		$v = ".$v";
	    }
	    $self->{cachefile} = "$v.cache";
	}
    }
    
    if (defined($v = delete $_{rw})) {
	$self->{rw} = $v;
    }
    
    if (keys(%_)) {
	foreach my $k (keys %_) {
	    carp "unknown parameter $k"
	}
	++$err;
    }
    return undef if $err;
    return $self;
}

sub DESTROY {
    my $self = shift;
    $self->writecache();
}

=head2 $cfg->error($message)

=head2 $cfg->error($message, locus => $loc)

Prints the B<$message> on STDERR.  If <locus> is given, its value must
be a reference to a valid B<App::Beam::Config::Locus>(3) object.  In that
case, the object will be formatted first, then followed by a ": " and the
B<$message>.    
    
=cut
    
sub error {
    my $self = shift;
    my $err = shift;
    local %_ = @_;
    $err = $_{locus}->format($err) if exists $_{locus};
    print STDERR "$err\n";
}

=head2 $cfg->debug($lev, @msg)

If B<$lev> is greater than or equal to the B<debug> value used when
creating B<$cfg>, outputs on standard error the strings from @msg,
separating them with a single space character.

Otherwise, does nothing.    

=cut    

sub debug {
    my $self = shift;
    my $lev = shift;
    return unless $self->{debug} >= $lev;
    $self->error("DEBUG: " . join(' ', @_));
}

sub writecache {
    my $self = shift;
    return unless exists $self->{cachefile};
    return unless exists $self->{conf};
    return unless $self->{updated};
    $self->debug(1, "storing cache file $self->{cachefile}");
    store $self->{conf}, $self->{cachefile};
}

sub parse_section {
    my ($self, $conf, $input) = @_;
    my $ref = $conf;
    my $quote;
    my $rootname;
    my $kw = $self->{parameters} if exists $self->{parameters};
    while ($input ne '') {
	my $name;
	if (!defined($quote)) {
	    if ($input =~ /^"(.*)/) {
		$quote = '';
		$input = $1;
	    } elsif ($input =~ /^(.+?)(?:\s+|")(.*)/) {
		$name = $1;
		$input = $2;
	    } else {
		$name = $input;
		$input = '';
	    }
	} else {
	    if ($input =~ /^([^\\"]*)\\(.)(.*)/) {
		$quote .= $1 . $2;
		$input = $3;
	    } elsif ($input =~ /^([^\\"]*)"\s*(.*)/) {
		$name = $quote . $1;
		$input = $2;
		$quote = undef;
	    } else {
		croak "unparsable input $input";
	    }
	}

	if (defined($name)) {
	    $rootname = $name unless defined $rootname;
	    $ref->{$name} = {} unless ref($ref->{$name}) eq 'HASH';
	    $ref = $ref->{$name};
	    
	    if (defined($kw)
		and ref($kw) eq 'HASH' and exists($kw->{$name}{section})) {
		$kw = $kw->{$name}{section};
	    }
	    
	    $name = undef;
	}
    }
    return ($ref, $rootname, $kw);
}

sub check_mandatory {
    my $self = shift;
    my $kw = shift;
    my $section = shift;
    my $loc = shift;
    
    my $err = 0;
    while (my ($k, $d) = each %{$kw}) {
	if (ref($d) eq 'HASH') {
	    if ($d->{mandatory} && !exists($section->{$k})) {
		$loc = $section->{-locus} if exists($section->{-locus});
		if (exists($d->{section})) {
		    $self->error("mandatory section ["
				 . join(' ', @_, $k)
				 . "] not present",
				 locus => $loc);
		    ++$err;
		} else {
		    $self->error("mandatory variable \""
				 . join('.', @_, $k)
				 . "\" not set",
				 locus => $loc);
		    ++$err;
		}
	    }
	    if (exists($d->{section})) {
		if ($k eq '*') {
		    while (my ($name, $vref) = each %{$section}) {
			if (is_section_ref($vref)) {
			    $err += $self->check_mandatory($d->{section},
							   $vref,
							   $loc,
							   @_, $name);
			}
		    }
		} elsif (exists($section->{$k})) {
		    $err += $self->check_mandatory($d->{section},
						   $section->{$k},
						   $loc,
						   @_, $k);
		}
	    }
	}
    }
    return $err;
}

sub readconfig {
    my $self = shift;
    my $file = shift;
    my $conf = shift;
    local %_ = @_;
    
    $self->debug(1, "reading file $file");
    open(my $fd, "<", $file)
	or do {
	    $self->error("can't open configuration file $file: $!");
	    return 1;
        };
    
    my $line;
    my $err;
    my $section = $conf;
    my $kw = $self->{parameters};
    my $include = 0;
    my $rootname;
    
    while (<$fd>) {
	++$line;
	chomp;
	if (/\\$/) {
	    chop;
	    $_ .= <$fd>;
	    redo;
	}
	
	s/^\s+//;
	s/\s+$//;
	s/#.*//;
	next if ($_ eq "");
	    
	if (/^\[(.+?)\]$/) {
	    $include = 0;
	    my $arg = $1;
	    $arg =~ s/^\s+//;
	    $arg =~ s/\s+$//;
	    if ($arg eq 'include') {
		$include = 1;
	    } else {
		($section, $rootname, $kw) = $self->parse_section($conf, $1);
		$self->error("unknown section",
			     locus => new App::Beam::Config::Locus($file, $line))
		    if (exists($self->{parameters}) and !defined($kw));
		if ($self->{locations}) {
		    $section->{-locus} =
			new App::Beam::Config::Locus($file, $line)
			    unless exists $section->{-locus};
		}
	    }
	} elsif (/([\w_-]+)\s*=\s*(.*)/) {
	    my ($k, $v) = ($1, $2);
	    $k = lc($k) if $self->{ci};

	    if ($include) {
		if ($k eq 'path') {
		    $err += $self->readconfig($v, $conf, include => 1);
		} elsif ($k eq 'pathopt') {
		    $err += $self->readconfig($v, $conf, include => 1)
			if -f $v;
		} elsif ($k eq 'glob') {
		    foreach my $file (bsd_glob($v, 0)) {
			$err += $self->readconfig($file, $conf, include => 1);
		    }
		} else {
		    $self->error("keyword \"$k\" is unknown",
				 locus => new App::Beam::Config::Locus($file, $line));
		    ++$err;
		}
		next;
	    }

	    if (defined($kw)) {
		my $x = $kw->{$k};
		$x = $kw->{'*'} unless defined $x;
		if (!defined($x)) {
		    $self->error("keyword \"$k\" is unknown",
				 locus => new App::Beam::Config::Locus($file, $line));
		    ++$err;
		    next;
		} elsif (ref($x) eq 'HASH') {
		    my $errstr;
		    my $prev_val;
		    if (exists($section->{$k})) {
			$prev_val = $section->{$k};
			$prev_val = $prev_val->{-value}
			    if ref($prev_val) eq 'HASH'
				&& exists($prev_val->{-value});
		    }
		    if (exists($x->{re})) {
			if ($v !~ /$x->{re}/) {
			    $self->error("invalid value for $k",
					 locus => new App::Beam::Config::Locus($file, $line));
			    ++$err;
			    next;
			}
		    }

		    if (exists($x->{check})) {
			if (defined($errstr = &{$x->{check}}(\$v, $prev_val))) {
			    $self->error($errstr,
					 locus => new App::Beam::Config::Locus($file, $line));
			    ++$err;
			    next;
			}
		    }

		    if ($x->{array}) {
			if (!defined($prev_val)) {
			    $v = [ $v ];
			} else {
			    $v = [ @{$prev_val}, $v ];
			}
		    }
		}
	    }
	    if ($self->{locations}) {
		$section->{-locus}->add($file, $line);
		unless (exists($section->{$k})) {
		    $section->{$k}{-locus} =
			new App::Beam::Config::Locus();
		}
		$section->{$k}{-locus}->add($file, $line);
		$section->{$k}{-value} = $v;
	    } else {
		$section->{$k} = $v;
	    }
        } else {
    	    $self->error("malformed line",
			 locus => new App::Beam::Config::Locus($file, $line));
	    ++$err;
	    next;
	}
    }
    close $fd;
    $err += $self->check_mandatory($self->{parameters}, $conf,
				   new App::Beam::Config::Locus($file, $line))
	unless $_{include};
    return $err;
}

sub fixup {
    my $self = shift;
    my $params = shift;
    while (my ($kv, $descr) = each %$params) {
	next unless ref($descr) eq 'HASH';
	if (exists($descr->{section})) {
	    $self->fixup($descr->{section}, @_, $kv);
	} elsif (exists($descr->{default}) && !$self->isset(@_, $kv)) {
	    $self->set(@_, $kv, $descr->{default});
	}
    }
}

sub file_up_to_date {
    my ($self, $file) = @_;
    my $st_conf = stat($self->{filename}) or return 1;
    my $st_file = stat($file)
	or carp "can't stat $file: $!";
    return $st_conf->mtime <= $st_file->mtime;
}

=head2 $cfg->parse()

Parses the configuration file and stores the data in the object.  Returns
true on success and false on failure.  Eventual errors in the configuration
are reported using B<$cfg->error>.

=cut

sub parse {
    my ($self) = @_;
    my %conf;

    if (exists($self->{cachefile}) and -f $self->{cachefile}) {
	if ($self->file_up_to_date($self->{cachefile})) {
	    my $ref;
	    $self->debug(1, "reading from cache file $self->{cachefile}");
	    eval { $ref = retrieve($self->{cachefile}); };
	    if (defined($ref)) {
		$self->{conf} = $ref;
		$self->{updated} = $self->{rw};
		return 1;
	    } elsif ($@) {
		$self->error("warning: unable to load configuration cache: $@");
	    }
	}
	unlink $self->{cachefile};
    }
    
    $self->debug(1, "parsing $self->{filename}");
    my $err = $self->readconfig($self->{filename}, \%conf);
    if ($err == 0) {
	$self->{conf} = \%conf ;
	$self->{updated} = 1;
	$self->fixup($self->{parameters}) if exists $self->{parameters};
    }
    return !$err;
}

sub getref {
    my $self = shift;
    
    return undef unless exists $self->{conf};
    my $ref = $self->{conf};
    for (@_) {
	return undef unless exists $ref->{$_};
	$ref = $ref->{$_};
    }
    return $ref;
}

=head2 $var = $cfg->get(@path);

Returns the value of the configuration variable represented by its
I<path>, or B<undef> if the variable is not set.  The path is a list
of configuration variables leading to the value in question.  For example,
the following statement:

    pidfile = /var/run/x.pid

has the path

    ( 'pidfile' )

The path of the B<pidfile> statement in section B<core>, e.g.:

    [core]
        pidfile = /var/run/x.pid

is

    ( 'core', 'pidfile' )

Similarly, the path of the B<file> setting in the following configuration
file:    

    [item foo]
        file = bar
    
is
    ( 'item', 'foo', 'bar' )
    
=head2 $ret = $cfg->get({ variable => $pathref, return => all | value | locus })

I<$pathref> is a reference to the configuration setting path as described
above.  This invocation is similar to B<get(@{$pathref})>, except that
it returns additional data as controlled by the B<return> keyword.  The
valid values for the B<return> are:

=over 4

=item 'value'

Returns the value of the variable.  The call

    $cfg->get({ variable => \@path, return => 'value' })

is completely equivalent to

    $cfg->get(@path);

=item 'locus'

If B<$cfg> was created with B<locations> enabled, returns the source
location of this configuration setting (see B<App::Beam::Config::Locus>(3)).

=item 'all'

Returns a reference to a hash with the following keys: B<-value> and B<-locus>.
The B<$ret{-value}> contains the value of the setting.  The B<$ret{-locus}>
is defined only if B<$cfg> was created with B<locations> enabled, and contains
a reference to B<App::Beam::Config::Locus>(3) describing the source location
where the setting was defined.    
    
=back

If the B<return> key is absent, the result is the same as for
return => 'all'.    

=cut    

sub get {
    my $self = shift;
    croak "no variable to get" unless @_;
    my $ref;
    my $ctl;
    if (ref($_[0]) eq 'HASH') {
	$ctl = shift;
	croak "too many arguments" if @_;
	croak "no variable to get" unless exists $ctl->{variable};
	$ref = $self->getref(@{$ctl->{variable}});
	if (defined($ref)
	    && exists($ctl->{return})
	    && $ctl->{return} ne 'all') {
	    if (exists($ref->{$ctl->{return}})) {
		$ref = $ref->{$ctl->{return}};
	    } else {
		$ref = undef;
	    }
	} elsif (!exists($ref->{-value})) {
	    $ref = { '-value' => $ref };
	} 
    } else {
	$ref = $self->getref(@_);
	if (defined($ref) && ref($ref) eq 'HASH' && exists($ref->{-value})) {
	    $ref = $ref->{-value};
	}
    }
    return $ref;
}

=head2 $cfg->isset(@path)

Returns true if the configuration variable addressed by B<@path> is
set.    
    
=cut

sub isset {
    my $self = shift;
    return defined $self->getref(@_);
}

sub is_section_ref {
    my ($ref) = @_;
    return ref($ref) eq 'HASH'
	   && !exists($ref->{-value});
}

=head2 $cfg->issection(@path)

Returns true if the configuration section addressed by B<@path> is
set.

=cut

sub issection {
    my $self = shift;
    my $ref = $self->getref(@_);
    return defined($ref) && is_section_ref($ref);
}

=head2 $cfg->isvariable(@path)

Returns true if the configuration variable addressed by B<@path> is
set.

=cut

sub isvariable {
    my $self = shift;
    my $ref = $self->getref(@_);
    return defined($ref) && !is_section_ref($ref);
}

=head2 $cfg->set(@path, $value)

Sets the configuration variable B<@path> to B<$value>.    

=cut

sub set {
    my $self = shift;
    $self->{conf} = {} unless exists $self->{conf};
    my $ref = $self->{conf};
   
    while ($#_ > 1) {
	my $arg = shift;
	$ref->{$arg} = {} unless exists $ref->{$arg};
	$ref = $ref->{$arg};
    }
    if ($self->{locations}) {
	$ref->{$_[0]}{-value} = $_[1];
    } else {
	$ref->{$_[0]} = $_[1];
    }
    $self->{updated} = $self->{rw};
}

=head2 cfg->unset(@path)

Unsets the configuration variable.
    
=cut

sub unset {
    my $self = shift;
    return unless exists $self->{conf};
    my $ref = $self->{conf};
    my @path;
    
    for (@_) {
	return unless exists $ref->{$_};
	push @path, [ $ref, $_ ];
	$ref = $ref->{$_};
    }

    while (1) {
	my $loc = pop @path;
	delete ${$loc->[0]}{$loc->[1]};
	last unless (defined($loc) and keys(%{$loc->[0]}) == 0);
    }
    $self->{updated} = $self->{rw};
}    

=head2 @array = $cfg->names_of(@path)

If B<@path> refers to an existing configuration section, returns a list
of names of variables and subsections defined within that section.  E.g.,
if you have

    [item foo]
       x = 1
    [item bar]
       x = 1
    [item baz]
       y = 2

the call

    $cfg->names_of('item')

will return

    ( 'foo', 'bar', 'baz' )
    
=cut    

sub names_of {
    my $self = shift;
    my $ref = $self->getref(@_);
    return () if !defined($ref) || ref($ref) ne 'HASH';
    return map { /^-/ ? () : $_ } keys %{$ref};
}

#sub each {
#    my $self = shift;
#    return @{[ each %{$self->{conf}} ]};
#}

sub _foreach {
    my ($self, $domain, $aref, $cb, $sort) = @_;
    my @keys = keys(%{$aref});
    foreach my $k (defined($sort) ? sort $sort @keys : @keys) {
	my $v = $aref->{$k};
	if (ref($v) eq 'HASH') {
	    if (exists($v->{-value})) {
		&{$cb}($k, $v->{-value}, @{$domain});
	    } else {
		push @{$domain}, $k;
		$self->_foreach($domain, $v, $cb, $sort);
		pop @{$domain};
	    }
	} elsif ($k !~ /^-/) {
	    &{$cb}($k, $v, @{$domain});
	}
    }
}

=head2 $cfg->foreach($callback, %opts)

For each configuration settings calls

    &{$callback}($keyword, $value, @path)

where B<$keyword> is the configuration keyword, B<$value> is its value, and
B<@path> is the path to the section within which the setting occurred (in
other words, the full path to the configuration keyword is
B<(@path, $keyword)>).

Only one option is defined: B<sort>.  If its value is B<1>, the keywords will
be passed to B<$callback> in lexicographical order.  It can also refer to
a user-defined sorting function.
    
=cut
    
sub foreach {
    my $self = shift;
    my $cb = shift;
    
    local %_ = @_;
    my $sort = delete $_{sort};

    if (defined($sort) && ref($sort) ne 'CODE') {
	if ($sort == 1) {
	    $sort = sub { $a cmp $b };
	} else {
	    carp "sort must refer to a CODE";
	    return;
	}
    }
    if (keys(%_)) {
	foreach my $k (keys %_) {
	    carp "unknown parameter $k";
	}
	return;
    }
	
    $self->_foreach([], $self->{conf}, $cb, $sort);
}

=head2 @array = $cfg->flatten()

Returns a I<flattened> representation of the configuration, as a
list of pairs B<[ $dottedpath, $value ]>, where B<$dottedpath> is
a dot-separated keyword path, and B<$value> is the value of that
setting.  E.g., the following configuration

    pidfile = /var/run/x.pid
    [item foo]
       name = bar
    [item foo bar]
       lock = On

is represented as:

    ('item.foo.name', 'bar')
    ('item.foo.bar.lock', 'On')
    
=cut

sub flatten {
    my $self = shift;
    my @dump;
    $self->foreach(sub {
	                my $k = shift;
			my $v = shift;
			push @_, $k;
			push @dump, [ join('.', @_), $v ];
		   },
		   sort => 1);
    return @dump;
}
